@@ -1,5 +1,5 @@
 #!/usr/bin/env python
-
+# encoding: utf8
 import wmldata, os, glob, sys
 import re
 
@@ -86,8 +86,6 @@ class Parser:
         # Whether to print current file, comments, and macro replacements.
         self.verbose = False
 
-        self.ignore_fatal_errors = False
-
     def read_encoded(self, filename):
         """
         Helper for gracefully handling non-utf8 files and fixing up non-unix
@@ -298,6 +296,19 @@ class Parser:
             text += line
         return text
 
+    def read_lua(self):
+        """
+        Read a << .... >> string, return contents literally.
+        """
+        text = ""
+        while not self.at_end():
+            c = self.read_next()
+            if c == ">" and self.peek_next() == ">":
+                self.read_next() #get rid of trailing >
+                return text
+            text += c
+        raise Error(self, "Unexpected end of file")
+
     def skip_whitespace_inside_statement(self):
         self.read_while(" \t\r\n")
         if not self.at_end():
@@ -374,10 +385,14 @@ class Parser:
                     return None
             # If it is a directory, parse all cfg files within.
             if os.path.isdir(dirpath):
+                # Note: glob.glob will try to return unicode filenames
+                # if you pass it an unicode string - but to deal with
+                # non-unicode filenames as are allowed in linux we
+                # convert to a byte-string.
                 # Execute all WML files in the directory.
-                files = glob.glob(dirpath + "/*.cfg")
+                files = glob.glob(str(dirpath + "/*.cfg"))
                 # And also execute directories with a _main.cfg.
-                files += glob.glob(dirpath + "/*/_main.cfg")
+                files += glob.glob(str(dirpath + "/*/_main.cfg"))
                 files.sort()
                 mc = dirpath + "/_main.cfg"
                 fc = dirpath + "/_final.cfg"
@@ -473,16 +488,9 @@ class Parser:
             text = macro.text
             for i in range(len(macro.params)):
                 if 1 + i >= len(params):
-                    if self.ignore_fatal_errors:
-                        sys.stderr.write("***TRYING TO IGNORE FATAL ERROR ***\n")
-                        sys.stderr.write("Not enough parameters for macro %s. " % name +
-                            "%d given but %d needed %s.\n" % (len(params) - 1,
-                                len(macro.params), macro.params))
-                        params.append("")
-                    else:
-                        raise Error(self, "Not enough parameters for macro %s. " % name +
-                            "%d given but %d needed %s." % (len(params) - 1,
-                                len(macro.params), macro.params))
+                    raise Error(self, "Not enough parameters for macro %s. " % name +
+                        "%d given but %d needed %s." % (len(params) - 1,
+                            len(macro.params), macro.params))
                 rep = params[1 + i]
                 # Handle gettext replacement here, since inside the macro
                 # the textdomain will be wrong.
@@ -498,9 +506,6 @@ class Parser:
                 text = text.replace("{%s}" % macro.params[i], rep)
 
             if text:
-                # Hack: Ignore empty lines at beginning of a macro at least one
-                # addon breaks otherwise.
-                while len(text) >= 1 and text[0] == '\n': text = text[1:]
                 self.push_text(name, text, initial_textdomain = macro.textdomain)
             else:
                 pass # empty macro, nothing to do
@@ -546,6 +551,7 @@ class Parser:
         spaces = ""
         filename = "(None)"
         line = -1
+        got_lua = False
         while 1:
             if c == "{":
                 keep_macro = self.parse_macro()
@@ -557,6 +563,13 @@ class Parser:
             elif c == "\n":
                 break
             elif c == "#":
+                # FIXME
+                # Assume something like:
+                # name="picture.png="+
+                # #textdomain blah
+                # "translateable name"
+                #
+                # For now, we ignore the textdomain...
                 self.read_until("\n")
                 break
             elif c == "+":
@@ -577,7 +590,10 @@ class Parser:
                     else:
                         variable += c  
             else:
-                if c == '"':
+                if c == "<" and self.peek_next() == "<":
+                    self.read_next() #skip the rest of the opening
+                    value += self.read_lua()
+                elif c == '"':
                     # We want the textdomain at the beginning of the string,
                     # the end of the string may be outside a macro and already
                     # in another textdomain.
@@ -637,6 +653,8 @@ class Parser:
         while 1:
             self.skip_whitespace_and_newlines()
             if self.at_end():
+                if state:
+                    raise Error(self, "Tag stack non-empty (%s) at end of data" % state)
                 break
             c = self.read_next()
             if c == "#": # comment or preprocessor
@@ -662,9 +680,13 @@ class Parser:
 
                 elif self.check_for("undef "):
                     self.read_until(" ")
-                    name = self.read_until(" \n")
+                    name = self.read_until("\n")
                     name = name.rstrip()
+                    if " " in name:
+                        if self.verbose: sys.stderr.write("Stray symbols in #undef %s\n" % name)
+                        name = name.split(" ")[0]
                     if name in self.macros: del self.macros[name]
+                    elif self.verbose: sys.stderr.write("undef'd macro '%s' did not exist\n" % name)
                 elif self.check_for("ifdef ") or self.check_for("ifndef"):
 
                     what = "#" + self.read_until(" ").rstrip()
@@ -758,22 +780,27 @@ class Parser:
                 if name[0] == '/':
                     if state == name[1:]:
                         return
-                    if self.ignore_fatal_errors:
-                        # This is just a hack to get broken addons to still show
-                        # up with wmlunits.
-                        sys.stderr.write("***TRYING TO IGNORE FATAL ERROR ***\n")
-                        sys.stderr.write("Mismatched closing tag [%s], expected [/%s]\n" % (name, state))
-                        continue
-                    else:
-                        raise Error(self, "Mismatched closing tag [%s], expected [/%s]" % (name, state))
-                #TODO: make a [+tag] properly append to most recent [tag]
-                # The below is an ugly hack, but better than keeping the '+' in the tag name.
+                    raise Error(self, "Mismatched closing tag [%s], expected [/%s]" % (name, state))
                 elif name[0] == '+':
                     name = name[1:]
-                subdata = wmldata.DataSub(name)
-                subdata.set_meta(self.filename, self.line)
-                self.parse_top(subdata, name)
-                data.insert(subdata)
+                    try:
+                        subdata = data.dict[name][-1]
+                        self.parse_top(subdata, name)
+                    except KeyError:
+                        subdata = wmldata.DataSub(name)
+                        subdata.set_meta(self.filename, self.line)
+                        self.parse_top(subdata, name)
+                        data.insert(subdata)
+                elif name[0] == " ":
+                    # We single this case out explicitly as the wesnoth parser
+                    # ignores it due to implementation reasons and this makes
+                    # more sense of the error message
+                    raise Error(self, "Invalid tag [%s] contains whitespace" % name)
+                else:
+                    subdata = wmldata.DataSub(name)
+                    subdata.set_meta(self.filename, self.line)
+                    self.parse_top(subdata, name)
+                    data.insert(subdata)
             elif c == '{':
                 keep_macro = self.parse_macro()
                 if isinstance(keep_macro, wmldata.Data):
@@ -783,12 +810,47 @@ class Parser:
                     data.insert(subdata)
 
 
+def strify(string):
+    """
+Massage a string into what appears to be a JSON-compatible form
+"""
+    s = repr(string)
+    front = s.index("'")
+    return '"%s"' % s[front+1:len(s)-1]
+
+def jsonify(tree, verbose=False):
+    """
+Convert a DataSub into JSON
+
+If verbose, insert a linebreak after every brace and comma (put every item on its own line), otherwise, condense everything into a single line.
+"""
+    print "{",
+    if verbose: print
+    first = True
+    for child in tree.children():
+        if first:
+            first = False
+        else:
+            print ",",
+            if verbose: print
+        print '"%s":' % child.name,
+        if child.get_type() == "DataSub":
+            jsonify(child)
+        else:
+            print strify(child.get_value()),
+            if verbose: print
+    print "}",
+    if verbose: print
+
 if __name__ == "__main__":
     import optparse, subprocess
     try: import psyco
     except ImportError: pass
     else: psyco.full()
 
+    # Hack to make us not crash when we encounter characters that aren't ASCII
+    sys.stdout = __import__("codecs").getwriter('utf-8')(sys.stdout)
+
     optionparser = optparse.OptionParser()
     optionparser.set_usage("usage: %prog [options] [filename]")
     optionparser.add_option("-p", "--path",  help = "specify wesnoth data path")
@@ -802,6 +864,8 @@ if __name__ == "__main__":
         help = "do not expand any macros")
     optionparser.add_option("-c", "--contents", action = "store_true",
         help = "display contents of every tag")
+    optionparser.add_option("-j", "--to-json", action = "store_true",
+        help = "output JSON version of tree")
     options, args = optionparser.parse_args()
 
     if options.path:
@@ -838,5 +902,8 @@ if __name__ == "__main__":
     data = wmldata.DataSub("WML")
     wmlparser.parse_top(data)
 
-    data.debug(show_contents = options.contents, use_color = options.color)
+    if options.to_json:
+        jsonify(data)#, True) # For more readable results
+    else:
+        data.debug(show_contents = options.contents, use_color = options.color)
 
